#include "Waves'NES.h"

#ifndef _MAPPER004_H_
#define _MAPPER004_H_

//////////////////////////////////////////////////////////////////////////
// Mapper004  Nintendo MMC3                                             //
//////////////////////////////////////////////////////////////////////////
#define	MMC3_IRQ_KLAX		1
#define	MMC3_IRQ_SHOUGIMEIKAN	2
#define	MMC3_IRQ_DAI2JISUPER	3
#define	MMC3_IRQ_DBZ2		4
#define	MMC3_IRQ_ROCKMAN3	5

	char	regd[8];
	char    XRAM[0x1000];
	BYTE	prg0, prg1;
	BYTE	chr01, chr23, chr4, chr5, chr6, chr7;
	BYTE	we_sram;

	BYTE	irq_type;
	BYTE	irq_enable;
	BYTE	irq_counter;
	BYTE	irq_latch;
	BYTE	irq_request;
	BYTE	irq_preset;
	BYTE	irq_preset_vbl;

	BYTE	vs_patch;
	BYTE	vs_index;

void	SetBank_CPU()
{
	if( regd[0] & 0x40 ) {
		SetPROM_32K_Bank2( PROM_8K_SIZE-2, prg1, prg0, PROM_8K_SIZE-1 );
	} else {
		SetPROM_32K_Bank2( prg0, prg1, PROM_8K_SIZE-2, PROM_8K_SIZE-1 );
	}
}

void	SetBank_PPU()
{
	if( VROM_1K_SIZE > 0 ) {
		if( regd[0] & 0x80 ) {
			SetVROM_8K_Bank1( chr4, chr5, chr6, chr7,
					 chr01, chr01+1, chr23, chr23+1 );
		} else {
			SetVROM_8K_Bank1( chr01, chr01+1, chr23, chr23+1,
					 chr4, chr5, chr6, chr7 );
		}
	} else {
		if( regd[0] & 0x80 ) {
			SetCRAM_1K_Bank( 4, (chr01+0)&0x07 );
			SetCRAM_1K_Bank( 5, (chr01+1)&0x07 );
			SetCRAM_1K_Bank( 6, (chr23+0)&0x07 );
			SetCRAM_1K_Bank( 7, (chr23+1)&0x07 );
			SetCRAM_1K_Bank( 0, chr4&0x07 );
			SetCRAM_1K_Bank( 1, chr5&0x07 );
			SetCRAM_1K_Bank( 2, chr6&0x07 );
			SetCRAM_1K_Bank( 3, chr7&0x07 );
		} else {
			SetCRAM_1K_Bank( 0, (chr01+0)&0x07 );
			SetCRAM_1K_Bank( 1, (chr01+1)&0x07 );
			SetCRAM_1K_Bank( 2, (chr23+0)&0x07 );
			SetCRAM_1K_Bank( 3, (chr23+1)&0x07 );
			SetCRAM_1K_Bank( 4, chr4&0x07 );
			SetCRAM_1K_Bank( 5, chr5&0x07 );
			SetCRAM_1K_Bank( 6, chr6&0x07 );
			SetCRAM_1K_Bank( 7, chr7&0x07 );
		}
	}
}

void   Mapper004_Reset()
{
int i;
	for(i = 0; i < 8; i++ ) {
		regd[i] = 0x00;
	}

	prg0 = 0;
	prg1 = 1;
	SetBank_CPU();

	chr01 = 0;
	chr23 = 2;
	chr4  = 4;
	chr5  = 5;
	chr6  = 6;
	chr7  = 7;
	SetBank_PPU();

	we_sram  = 0;	// Disable
	irq_enable = 0;	// Disable
	irq_counter = 0;
	irq_latch = 0xFF;
	irq_request = 0;
	irq_preset = 0;
	irq_preset_vbl = 0;

	// IRQ?????
	//nes->SetIrqType( NES::IRQ_CLOCK );
	irq_type = 0;
}

BYTE	Mapper004_ReadLow( WORD addr )
{
	if( !vs_patch ) {
		if( addr >= 0x5000 && addr <= 0x5FFF ) {
			return	XRAM[addr-0x4000];
		}
	} else if( vs_patch == 1 ) {
		// VS TKO Boxing Security
		if( addr == 0x5E00 ) {
			vs_index = 0;
			return	0x00;
		} else if( addr == 0x5E01 ) {
			BYTE	VS_TKO_Security[32] = {
				0xff, 0xbf, 0xb7, 0x97, 0x97, 0x17, 0x57, 0x4f,
				0x6f, 0x6b, 0xeb, 0xa9, 0xb1, 0x90, 0x94, 0x14,
				0x56, 0x4e, 0x6f, 0x6b, 0xeb, 0xa9, 0xb1, 0x90,
				0xd4, 0x5c, 0x3e, 0x26, 0x87, 0x83, 0x13, 0x00 };
			return	VS_TKO_Security[(vs_index++) & 0x1F];
		}
	} else if( vs_patch == 2 ) {
		// VS Atari RBI Baseball Security
		if( addr == 0x5E00 ) {
			vs_index = 0;
			return	0x00;
		} else if( addr == 0x5E01 ) {
			if( vs_index++ == 9 )
				return	0x6F;
			else
				return	0xB4;
		}
	} else if( vs_patch == 3 ) {
		// VS Super Xevious
		switch( addr ) {
			case	0x54FF:
				return	0x05;
			case	0x5678:
				if( vs_index )
					return	0x00;
				else
					return	0x01;
				break;
			case	0x578f:
				if( vs_index )
					return	0xD1;
				else
					return	0x89;
				break;
			case	0x5567:
				if( vs_index ) {
					vs_index = 0;
					return	0x3E;
				} else {
					vs_index = 1;
					return	0x37;
				}
				break;
			default:
				break;
		}
	}

	return	Mapper004_ReadLow( addr );
}

void	Mapper004_WriteLow( WORD addr, BYTE str )
{
	if( addr >= 0x5000 && addr <= 0x5FFF ) {
		XRAM[addr-0x4000] = str;
	} else {
		Mapper004_WriteLow( addr, str );
	}
}

void	Mapper004_Write( WORD addr, BYTE str )
{
	switch( addr & 0xE001 ) {
		case	0x8000:
			regd[0] = str;
			SetBank_CPU();
			SetBank_PPU();
			break;
		case	0x8001:
			regd[1] = str;

			switch( regd[0] & 0x07 ) {
				case	0x00:
					chr01 = str & 0xFE;
					SetBank_PPU();
					break;
				case	0x01:
					chr23 = str & 0xFE;
					SetBank_PPU();
					break;
				case	0x02:
					chr4 = str;
					SetBank_PPU();
					break;
				case	0x03:
					chr5 = str;
					SetBank_PPU();
					break;
				case	0x04:
					chr6 = str;
					SetBank_PPU();
					break;
				case	0x05:
					chr7 = str;
					SetBank_PPU();
					break;
				case	0x06:
					prg0 = str;
					SetBank_CPU();
					break;
				case	0x07:
					prg1 = str;
					SetBank_CPU();
					break;
			}
			break;
		case	0xA000:
			regd[2] = str;
			if( (RomHeader[6]&0x01 | RomHeader[6]&0x08)!=0x03 ) {
				if( str & 0x01 ) SetVRAM_Mirror( VRAM_HMIRROR );
				else		  SetVRAM_Mirror( VRAM_VMIRROR );
			}
			break;
		case	0xA001:
			regd[3] = str;
			break;
		case	0xC000:
			regd[4] = str;
			if( irq_type == MMC3_IRQ_KLAX || irq_type == MMC3_IRQ_ROCKMAN3 ) {
				irq_counter = str;
			} else {
				irq_latch = str;
			}
			if( irq_type == MMC3_IRQ_DBZ2 ) {
				irq_latch = 0x07;
			}
			break;
		case	0xC001:
			regd[5] = str;
			if( irq_type == MMC3_IRQ_KLAX || irq_type == MMC3_IRQ_ROCKMAN3 ) {
				irq_latch = str;
			} else {
				if( (irq_type == MMC3_IRQ_SHOUGIMEIKAN) ) {
					irq_counter |= 0x80;
					irq_preset = 0xFF;
				} else {
					irq_counter |= 0x80;
					irq_preset_vbl = 0xFF;
					irq_preset = 0;
				}
			}
			break;
		case	0xE000:
			regd[6] = str;
			irq_enable = 0;
			irq_request = 0;

			//nes->cpu->ClrIRQ( IRQ_MAPPER );
			break;
		case	0xE001:
			regd[7] = str;
			irq_enable = 1;
			irq_request = 0;

//			nes->cpu->ClrIRQ( IRQ_MAPPER );
			break;
	}
}

void	Mapper004_Clock( INT cycles )
{
//	if( irq_request && (nes->GetIrqType() == NES::IRQ_CLOCK) ) {
//		nes->cpu->IRQ_NotPending();
//	}
}

void	Mapper004_HSync( INT scanline )
{
	if( irq_type == MMC3_IRQ_KLAX ) {
		if( (scanline >= 0 && scanline <= 239)/* && nes->ppu->IsDispON()*/ ) {
			if( irq_enable ) {
				if( irq_counter == 0 ) {
					irq_counter = irq_latch;
					irq_request = 0xFF;
				}
				if( irq_counter > 0 ) {
					irq_counter--;
				}
			}
		}
		if( irq_request ) {
			//nes->cpu->SetIRQ( IRQ_MAPPER );
		}
	} else if( irq_type == MMC3_IRQ_ROCKMAN3 ) {
		if( (scanline >= 0 && scanline <= 239) /*&& nes->ppu->IsDispON()*/ ) {
			if( irq_enable ) {
				if( !(--irq_counter) ) {
					irq_request = 0xFF;
					irq_counter = irq_latch;
				}
			}
		}
		if( irq_request ) {
			//nes->cpu->SetIRQ( IRQ_MAPPER );
		}
	} else {
		if( (scanline >= 0 && scanline <= 239)/* && nes->ppu->IsDispON()*/ ) {
			if( irq_preset_vbl ) {
				irq_counter = irq_latch;
				irq_preset_vbl = 0;
			}
			if( irq_preset ) {
				irq_counter = irq_latch;
				irq_preset = 0;
				if( irq_type == MMC3_IRQ_DAI2JISUPER && scanline == 0 ) {
					irq_counter--;
				}
			} else if( irq_counter > 0 ) {
				irq_counter--;
			}

			if( irq_counter == 0 ) {
				if( irq_enable ) {
					irq_request = 0xFF;

					//nes->cpu->SetIRQ( IRQ_MAPPER );

#if	0
{
LPBYTE	lpScn = nes->ppu->GetScreenPtr();

	lpScn = lpScn+(256+16)*nes->GetScanline();

	for( INT i = 0; i < 16; i++ ) {
		lpScn[i] = 22;
	}
}
#endif
				}
				irq_preset = 0xFF;
			}
		}
	}
//	if( irq_request && (nes->GetIrqType() == NES::IRQ_HSYNC) ) {
//		nes->cpu->IRQ_NotPending();
//	}
}

void	Mapper004_SaveState( char* p )
{
int i;
	for(i = 0; i < 8; i++ ) {
		p[i] = regd[i];
	}
	p[ 8] = prg0;
	p[ 9] = prg1;
	p[10] = chr01;
	p[11] = chr23;
	p[12] = chr4;
	p[13] = chr5;
	p[14] = chr6;
	p[15] = chr7;
	p[16] = irq_enable;
	p[17] = (BYTE)irq_counter;
	p[18] = irq_latch;
	p[19] = irq_request;
	p[20] = irq_preset;
	p[21] = irq_preset_vbl;
}

void	Mapper004_LoadState( char* p )
{
int i;
	for( i = 0; i < 8; i++ ) {
		regd[i] = p[i];
	}
	prg0  = p[ 8];
	prg1  = p[ 9];
	chr01 = p[10];
	chr23 = p[11];
	chr4  = p[12];
	chr5  = p[13];
	chr6  = p[14];
	chr7  = p[15];
	irq_enable  = p[16];
	irq_counter = (INT)p[17];
	irq_latch   = p[18];
	irq_request = p[19];
	irq_preset  = p[20];
	irq_preset_vbl = p[21];
}

void Mapper004_Init()
{
        NES_Mapper.Reset = Mapper004_Reset;
        NES_Mapper.Write = Mapper004_Write;
        NES_Mapper.WriteLow = Mapper004_WriteLow;
        NES_Mapper.ReadLow = Mapper004_ReadLow;
}

#endif