#include <string.h>
#include "Waves'NES.h"
#include "CPU.c"
#include "PPU.c"
#include "PPUINC.c"
#include "APU.c"
#include "Mapper.c"

#define PRG_BLOCK_SIZE        0x4000
#define PAT_BLOCK_SIZE        0x2000

sbit TFT_CS2 at GPIOE_ODR.B15;
 
int PROM_8K_SIZE;
int PROM_16K_SIZE;
int PROM_32K_SIZE;
BYTE *CPU_MEM_BANK[8];  // * 8192

int VROM_1K_SIZE;
int VROM_2K_SIZE;
int VROM_4K_SIZE;
int VROM_8K_SIZE;
BYTE *PPU_MEM_BANK[12];

BYTE CRAM[8*1024];//32
BYTE VRAM[4*1024];//4
BYTE RomHeader[20];
BYTE GRAM[0x800]; //800
BYTE SRAM[0x800]; //800

BYTE PRGBlock[4*PRG_BLOCK_SIZE];        // * 16384
BYTE PatternTable[4*PAT_BLOCK_SIZE];    // * 8192

BYTE NameTable[0x800];

BYTE BGPal[0x10];
BYTE SPPal[0x10];

BYTE SPRAM[0x100];
BYTE LineBit[256];
BYTE RevertByte[256];

BYTE CPUREG[0x18];                // Nes $4000-$4017

SPRITE* Sprite=(SPRITE*)SPRAM;

BYTE NES_buf[512];

void RUN_WaveNES(char *name_game)
{
        NES_Init();
        LoadRomFile(name_game);
        TFT_Fill_Screen(CL_BLACK);
        NES_Start();
        TFT_CS1 = 0;
        TFT_Set_Address_Ptr(32,0);
        TFT_CS1 = 1;
        NES_FrameExec();

}

void NES_Init()
{        
        int i,j;
        BYTE c,mask;
        for(i=0;i<256;i++)
        {
                c=0;
                mask=0x80;
                for(j=0;j<8;j++)if(i&(1<<j))c|=(mask>>j);
                RevertByte[i]=c;
        }
}

void NES_Start()
{
        memset(GRAM,      0, sizeof(GRAM));
        memset(SRAM,      0, sizeof(SRAM));
        memset(NameTable, 0, sizeof(NameTable));
        CPU_Reset();
        PPU_Reset();
}

void NES_FrameExec() //LineBit
{
        int i;
        char bt = 1;
        for(;;)
        {         /*
                  LCD->LCD_REG=R32;
                  LCD->LCD_RAM=0;
                  LCD->LCD_REG=R33;
                  LCD->LCD_RAM=30;
                LCD->LCD_REG=0x22;//ʼдRAM
                LED1=!LED1; 

                TFT_CS1 = 0;
                TFT_Set_Address_Ptr(32,0);
                TFT_CS1 = 1;
                */ 
                bt =~bt;
                for(i=0;i<240;i++)
                {
                        PPU_ScanLine(LineBit,i, 100, bt);
                        CPU_Exec(150);              //106

                        //GPIOB_ODR.B0 =~GPIOB_ODR.B0;
                        
                        if(PPU.m_REG[1]&0x18)
                        {
                                m_CurLineOft=m_ScreenOffset;
                                PPU.m_CurByteIndex=PPU.m_ByteIndex;
                        }
                }

                PPU.m_REG[2]|=PPU_VBLANK_FLAG;
                if(PPU.m_REG[0]&PPU_VBLANK_BIT) CPU_NMI();
                for(i=242;i<262;i++)
                {
                        CPU_ExecOnBaseCycle(1024);  //1024
                        CPU_ExecOnBaseCycle(338);   //338
                }
                PPU.m_REG[2]&=~PPU_VBLANK_FLAG;
                PPU.m_REG[2]&=~PPU_SPHIT_FLAG;
                Get_key();
        }         

}

void LoadRomFile(char *NES_ROM)
{
        int i;
        short err;
        int MapperNo;
        __HANDLE File_game;
        char lcd_buff[128];
        
        memset(CPU_MEM_BANK,0,sizeof(CPU_MEM_BANK));
        memset(PPU_MEM_BANK,0,sizeof(PPU_MEM_BANK));
        memset(RomHeader,0,sizeof(RomHeader));
        memset(CRAM,0,sizeof(CRAM));
        memset(VRAM,0,sizeof(VRAM));
        memset(CPUREG,0,sizeof(CPUREG));
        memset(APUREG,0,sizeof(APUREG));
        
        if(File_game = FAT32_Open(NES_ROM, FILE_READ)!=0)  TFT_Write_Text("Error file",10,10);
        
        TFT_Set_Font(&TFT_defaultFont, CL_GREEN, FO_HORIZONTAL);
        sprintf(lcd_buff, "File Name: %s", NES_ROM);
        TFT_Write_Text(lcd_buff,10,10);
        
        FAT32_Read(File_game, &RomHeader, 16);                                         //  16

        if((RomHeader[6] & 0x04))FAT32_Seek(File_game,512);

        if(RomHeader[4] > 0)
        {
        for(i=0; i<RomHeader[4];i++){
        FAT32_Read(File_game, PRGBlock + i*PRG_BLOCK_SIZE, PRG_BLOCK_SIZE);
          }
        }
        
        if(RomHeader[5] > 0)
        {
        for(i=0; i<RomHeader[5];i++){
        FAT32_Read(File_game, PatternTable + i*PAT_BLOCK_SIZE, PAT_BLOCK_SIZE);
          }
        }

        PROM_8K_SIZE=(uint8)RomHeader[4]*2;
        PROM_16K_SIZE=(uint8)RomHeader[4];
        PROM_32K_SIZE=(uint8)RomHeader[4]/2;
        
        VROM_1K_SIZE=(uint8)RomHeader[5]*8;
        VROM_2K_SIZE=(uint8)RomHeader[5]*4;
        VROM_4K_SIZE=(uint8)RomHeader[5]*2;
        VROM_8K_SIZE=(uint8)RomHeader[5];
        
        if(VROM_8K_SIZE){
        SetVROM_8K_Bank(0);
        }
        else{
        SetCRAM_8K_Bank(0);
        }
        
        if(RomHeader[6]&0x02){
        CPU_MEM_BANK[3] = SRAM;
        }
        
               if( RomHeader[6]&0x08 ) {
                SetVRAM_Mirror( VRAM_MIRROR4 );
                Mirror_type( VRAM_MIRROR4 );
        } else if( RomHeader[6]&0x01 ) {
                SetVRAM_Mirror( VRAM_VMIRROR );
                Mirror_type( VRAM_VMIRROR  );
        } else {
                SetVRAM_Mirror( VRAM_HMIRROR );
                Mirror_type( VRAM_HMIRROR );
        }
        
        if(RomHeader[7] == 0x44)MapperNo=(RomHeader[6] >> 4);
        else MapperNo = RomHeader[7] & 0xF0 | RomHeader[6] >> 4;

        CreateMapper(MapperNo);
        NES_Mapper.Reset();

        TFT_Set_Font(&TFT_defaultFont, CL_YELLOW, FO_HORIZONTAL);
        sprintf(lcd_buff, "Header: %c%c%c / %02X / %02X %02X / %02X %02X", RomHeader[0], RomHeader[1], RomHeader[2], RomHeader[3],
                                                                           RomHeader[4], RomHeader[5], RomHeader[6], RomHeader[7]);
        TFT_Write_Text(lcd_buff,10,30);
        TFT_Set_Font(&TFT_defaultFont, CL_GREEN, FO_HORIZONTAL);
        sprintf(lcd_buff, "CPU:    %02X %02X %02X %02X %02X %02X %02X %02X", CPU_MEM_BANK[0x0000>>13][0x0000&0x1FFF], CPU_MEM_BANK[0x2000>>13][0x2000&0x1FFF],CPU_MEM_BANK[0x4000>>13][0x4000&0x1FFF],CPU_MEM_BANK[0x6000>>13][0x6000&0x1FFF],CPU_MEM_BANK[0x8000>>13][0x8000&0x1FFF],CPU_MEM_BANK[0xA000>>13][0xA000&0x1FFF],CPU_MEM_BANK[0xC000>>13][0xC000&0x1FFF],CPU_MEM_BANK[0xE000>>13][0xE000&0x1FFF]);
        TFT_Write_Text(lcd_buff,10,45);
        sprintf(lcd_buff, "PPU:    %02X %02X %02X %02X %02X %02X %02X %02X", PPU_MEM_BANK[0x0000>>13][0x0000&0x1FFF], PPU_MEM_BANK[0x0400>>13][0x0400&0x1FFF],PPU_MEM_BANK[0x0800>>13][0x0800&0x1FFF],PPU_MEM_BANK[0x0C00>>13][0x0C00&0x1FFF],PPU_MEM_BANK[0x1000>>13][0x1000&0x1FFF],PPU_MEM_BANK[0x1400>>13][0x1400&0x1FFF],PPU_MEM_BANK[0x1800>>13][0x1800&0x1FFF],PPU_MEM_BANK[0x1C00>>13][0x1C00&0x1FFF]);
        TFT_Write_Text(lcd_buff,10,60);
        
        if(RomHeader[4] > 4)TFT_Set_Font(&TFT_defaultFont, CL_RED, FO_HORIZONTAL);
           else             TFT_Set_Font(&TFT_defaultFont, CL_GREEN, FO_HORIZONTAL);
           
        sprintf(lcd_buff, "PRG_ROM: %i x 16k",RomHeader[4]);
        TFT_Write_Text(lcd_buff,10,80);
        
        if(RomHeader[5] > 4)TFT_Set_Font(&TFT_defaultFont, CL_RED, FO_HORIZONTAL);
           else             TFT_Set_Font(&TFT_defaultFont, CL_GREEN, FO_HORIZONTAL);
        
        TFT_Set_Font(&TFT_defaultFont, CL_GREEN, FO_HORIZONTAL);
        sprintf(lcd_buff, "CHR_ROM: %i x 8k",RomHeader[5]);
        TFT_Write_Text(lcd_buff,10,95);
        
        sprintf(lcd_buff, "Mapper #: %i",MapperNo);
        TFT_Write_Text(lcd_buff,10,130);
        
        sprintf(lcd_buff, "Mirroring Type: %i    0x%02X",RomHeader[6]&0x01 | RomHeader[6]&0x08, RomHeader[6]&0x01 | RomHeader[6]&0x08);
        TFT_Write_Text(lcd_buff,10,170);
        /*
        sprintf(lcd_buff, "%02X %02X %02X %02X %02X %02X %02X %02X", CPU_ReadW(RES_VECTOR), CPU_MEM_BANK[0xA001>>13][0xA001&0x1FFF],CPU_MEM_BANK[2],CPU_MEM_BANK[3],CPU_MEM_BANK[4],CPU_MEM_BANK[5],CPU_MEM_BANK[6],CPU_MEM_BANK[7]);
        TFT_Write_Text(lcd_buff,10,200);
        */
        delay_ms(300);
        while(GPIOC_IDR.B13 == 1);
}
//--------------------------------------------PROM----------------------------------------------
void SetPROM_8K_Bank(BYTE page, int bank)                                     //0x8000 - 0xE0000
{
        bank %= PROM_8K_SIZE;
        CPU_MEM_BANK[page] = PRGBlock + 0x2000 * bank;
}

void SetPROM_16K_Bank(BYTE page, int bank)
{
        SetPROM_8K_Bank(page + 0, bank * 2 + 0);
        SetPROM_8K_Bank(page + 1, bank * 2 + 1);
}

void SetPROM_32K_Bank1(int bank)
{
        SetPROM_8K_Bank(4, bank * 4 + 0);
        SetPROM_8K_Bank(5, bank * 4 + 1);
        SetPROM_8K_Bank(6, bank * 4 + 2);
        SetPROM_8K_Bank(7, bank * 4 + 3);
}

void SetPROM_32K_Bank2(int bank0, int bank1, int bank2, int bank3)
{
        SetPROM_8K_Bank(4, bank0);   //0x8000
        SetPROM_8K_Bank(5, bank1);   //0xA000
        SetPROM_8K_Bank(6, bank2);   //0xC000
        SetPROM_8K_Bank(7, bank3);   //0xE000
}
//--------------------------------------------VROM----------------------------------------------
void SetVROM_1K_Bank(BYTE page, int bank)              // PPU[bank0-bank7]    //0x0000 - 0x1C000
{
        bank %= VROM_1K_SIZE;
        PPU_MEM_BANK[page] = PatternTable + 0x0400 * bank;
}

void SetVROM_2K_Bank(BYTE page, int bank)
{
        SetVROM_1K_Bank(page + 0, bank * 2 + 0);
        SetVROM_1K_Bank(page + 1, bank * 2 + 1);
}

void SetVROM_4K_Bank(BYTE page, int bank)
{
        SetVROM_1K_Bank(page + 0, bank * 4 + 0);
        SetVROM_1K_Bank(page + 1, bank * 4 + 1);
        SetVROM_1K_Bank(page + 2, bank * 4 + 2);
        SetVROM_1K_Bank(page + 3, bank * 4 + 3);
}

void SetVROM_8K_Bank(int bank)
{
        int i;
        for(i = 0; i < 8; i++){
        SetVROM_1K_Bank(i, bank * 8 + i);
        }
}

void SetVROM_8K_Bank1(int bank0, int bank1, int bank2, int bank3, int bank4, int bank5, int bank6, int bank7)
{
        SetVROM_1K_Bank(0, bank0 * 8 + 0);
        SetVROM_1K_Bank(1, bank1 * 8 + 1);
        SetVROM_1K_Bank(2, bank2 * 8 + 2);
        SetVROM_1K_Bank(3, bank3 * 8 + 3);
        SetVROM_1K_Bank(4, bank4 * 8 + 4);
        SetVROM_1K_Bank(5, bank5 * 8 + 5);
        SetVROM_1K_Bank(6, bank6 * 8 + 6);
        SetVROM_1K_Bank(7, bank7 * 8 + 7);
}

/*void SetVROM_8K_Bank1(int bank0, int bank1, int bank2, int bank3, int bank4, int bank5, int bank6, int bank7)
{
        SetVROM_1K_Bank(0, bank0);
        SetVROM_1K_Bank(1, bank1);
        SetVROM_1K_Bank(2, bank2);
        SetVROM_1K_Bank(3, bank3);
        SetVROM_1K_Bank(4, bank4);
        SetVROM_1K_Bank(5, bank5);
        SetVROM_1K_Bank(6, bank6);
        SetVROM_1K_Bank(7, bank7);
}*/
//--------------------------------------------CRAM----------------------------------------------
void SetCRAM_1K_Bank(BYTE page, int bank)                                     //0x0000 - 0x1C000???
{
        bank &= 0x1F;
        PPU_MEM_BANK[page] = CRAM + 0x0400 * bank;
}

void SetCRAM_2K_Bank(BYTE page, int bank)
{
        SetCRAM_1K_Bank(page + 0, bank * 2 + 0);
        SetCRAM_1K_Bank(page + 1, bank * 2 + 1);
}

void SetCRAM_4K_Bank(BYTE page, int bank)
{
        SetCRAM_1K_Bank(page + 0, bank * 4 + 0);
        SetCRAM_1K_Bank(page + 1, bank * 4 + 1);
        SetCRAM_1K_Bank(page + 2, bank * 4 + 2);
        SetCRAM_1K_Bank(page + 3, bank * 4 + 3);
}

void SetCRAM_8K_Bank(int bank)
{
        int i;
        for (i = 0; i < 8; i++)SetCRAM_1K_Bank(i, bank * 8 + i);
}
//--------------------------------------------VRAM----------------------------------------------
void SetVRAM_1K_Bank(BYTE page, int bank)                    //PPU            //0x2000 - 0x2C000
{
        bank &= 3;
        PPU_MEM_BANK[page] = VRAM + 0x0400 * bank;
}
//-----------------------------------------Table_Bank-------------------------------------------
void SetNameTable_Bank(int bank0, int bank1, int bank2, int bank3)            //0x2000 - 0x2C000
{
        SetVRAM_1K_Bank( 8, bank0);   //0x2000
        SetVRAM_1K_Bank( 9, bank1);   //0x2400
        SetVRAM_1K_Bank(10, bank2);   //0x2800
        SetVRAM_1K_Bank(11, bank3);   //0x2C00
}

void  SetVRAM_Mirror( INT type )
{
        switch( type ) {
                case    VRAM_HMIRROR:   // VRAM_HMIRROR
                        SetNameTable_Bank( 0, 0, 1, 1 );
                        break;
                case    VRAM_VMIRROR:   // VRAM_VMIRROR
                        SetNameTable_Bank( 0, 1, 0, 1 );  
                        break;
                case    VRAM_MIRROR4L:   // VRAM_MIRROR4L
                        SetNameTable_Bank( 0, 0, 0, 0 );
                        break;
                case    VRAM_MIRROR4H:   // VRAM_MIRROR4H
                        SetNameTable_Bank( 1, 1, 1, 1 );
                        break;
                case    VRAM_MIRROR4:   // VRAM_MIRROR4
                        SetNameTable_Bank( 0, 1, 2, 3 );
                        break;
        }
}

void  Mirror_type( INT type )
{
        switch( type ) {
                case    VRAM_HMIRROR:   // VRAM_HMIRROR
                        TFT_Write_Text("Mirroring: HMIRROR",10,145);
                        break;
                case    VRAM_VMIRROR:   // VRAM_VMIRROR
                        TFT_Write_Text("Mirroring: VMIRROR",10,145);
                        break;
                case    VRAM_MIRROR4L:   // VRAM_MIRROR4L
                        TFT_Write_Text("Mirroring: MIRROR4L",10,145);
                        break;
                case    VRAM_MIRROR4H:   // VRAM_MIRROR4H
                        TFT_Write_Text("Mirroring: MIRROR4H",10,145);
                        break;
                case    VRAM_MIRROR4:   // VRAM_MIRROR4
                        TFT_Write_Text("Mirroring: MIRROR4",10,145);
                        break;
        }
}

BYTE NES_ReadReg(WORD addr)
{
        switch(addr&0xFF)
        {
                case 0x00: case 0x01: case 0x02: case 0x03:
                case 0x04: case 0x05: case 0x06: case 0x07:
                case 0x08: case 0x09: case 0x0A: case 0x0B:
                case 0x0C: case 0x0D: case 0x0E: case 0x0F:
                case 0x10: case 0x11: case 0x12: case 0x13:
                          return APU_Read(addr);
                case 0x15:return APU_Read(addr);
                case 0x14:return addr&0xFF;  //14
                case 0x16:return (JOYPAD_GetValue(0)|0x40);
                case 0x17:return JOYPAD_GetValue(1)|APU_Read(addr);
                default:
                  return NES_Mapper.ExRead(addr);
                break;
        }
}

BYTE NES_Read(WORD addr)
{
        switch(addr>>13)
        {
                case 0x00:return GRAM[addr&0x07FF];
                case 0x01:return PPU_ReadFromPort(addr&0xE007);
                case 0x02:
                        if(addr<0x4100)return NES_ReadReg(addr);
                        else return NES_Mapper.ReadLow(addr);
                case 0x03:return NES_Mapper.ReadLow(addr);
                case 0x04:
                case 0x05:
                case 0x06:        
                case 0x07:return CPU_MEM_BANK[addr>>13][addr&0x1FFF];
        }
        return 0x00;
}

void NES_WriteReg(WORD addr, BYTE val)
{                
        switch(addr&0xFF)
        {
                case 0x00: case 0x01: case 0x02: case 0x03:
                case 0x04: case 0x05: case 0x06: case 0x07:
                case 0x08: case 0x09: case 0x0A: case 0x0B:
                case 0x0C: case 0x0D: case 0x0E: case 0x0F:
                case 0x10: case 0x11: case 0x12: case 0x13:
                case 0x15:
                        APU_Write(addr, val);
                        CPUREG[addr & 0xFF] = val;
                        break;
                case 0x14:
                        memcpy(Sprite,CPU_GetRAM(((WORD)val)<<8),256);
                        CPU.m_DMACycle+=514; //514
                        CPUREG[addr & 0xFF] = val;
                        break;
                case 0x16:
                        JOYPAD_InputBursh(val);
                        CPUREG[addr & 0xFF] = val;
                        break;
                 case 0x17:
                        CPUREG[addr & 0xFF] = val;
                        APU_Write(addr, val);
                        break;
                case 0x18:
                        APU_Write(addr, val);
                        break;
                default:
                  NES_Mapper.ExWrite(addr, val);
                break;
        }
}

void NES_Write(WORD addr, BYTE val)
{
        switch(addr>>13)
        {
                case 0x00:
                        GRAM[addr&0x07FF]=val;
                        break;
                case 0x01:
                        PPU_WriteToPort(addr&0xE007,val);
                        break;
                case 0x02:        
                        if(addr<0x4100)NES_WriteReg(addr,val);
                        else NES_Mapper.WriteLow(addr,val);
                        break;
                case 0x03:NES_Mapper.WriteLow(addr,val);break;
                case 0x04:
                case 0x05:
                case 0x06:
                case 0x07:        
                        NES_Mapper.Write(addr, val);
                        break;
        }
}

BYTE* NES_GetRAM(WORD addr)
{
        switch ( addr>>13 )
        {
                case 0x00:return &GRAM[addr&0x07FF];
                case 0x04:
                case 0x05:
                case 0x06:
                case 0x07:return &CPU_MEM_BANK[addr>>13][addr&0x1FFF];
        }
        return 0;
}