#ifndef _WAVESNES_H_
#define _WAVESNES_H_
#include "JOYPAD.h"
#include "mytype.h"

#define   PRG_BLOCK_SIZE        0x4000
#define   PAT_BLOCK_SIZE        0x2000
#define   FETCH_CYCLES               8

#define        VRAM_HMIRROR         0x00        // Horizontal
#define        VRAM_VMIRROR         0x01        // Virtical
#define        VRAM_MIRROR4         0x02        // All screen
#define        VRAM_MIRROR4L        0x03        // $2000-$23FF
#define        VRAM_MIRROR4H        0x04        // $2400-$27FF

static const unsigned int NES_Palette[64]=
{
        0x73AE,0x20D1,0x0015,0x4013,0x880E,0xA802,0xA000,0x7840,
        0x4160,0x0220,0x0280,0x01E2,0x19EB,0x0000,0x0000,0x0000,
        0xBDF7,0x039D,0x21DD,0x801E,0xB817,0xE00B,0xD940,0xCA61,
        0x8B80,0x04A0,0x0540,0x0487,0x0411,0x0000,0x0000,0x0000,
        0xFFFF,0x3DFF,0x5CBF,0xA45F,0xF3DF,0xFBB6,0xFBAC,0xFCC7,
        0xF5E7,0x8682,0x4EE9,0x5FD3,0x075B,0x0000,0x0000,0x0000,
        0xFFFF,0xAF3F,0xC6BF,0xD65F,0xFE3F,0xFE3B,0xFDF6,0xFED5,
        0xFF34,0xE7F4,0xAF97,0xB7F9,0x9FFE,0x0000,0x0000,0x0000
};

typedef struct tagNESCONFIG
{
        unsigned long BaseClock;                          // NTSC:21477270.0  PAL:21281364.0
        unsigned long CpuClock;                           // NTSC: 1789772.5  PAL: 1773447.0
        int           TotalScanlines;                     // NTSC: 262        PAL: 312
        int           ScanlineCycles;                     // NTSC:1364        PAL:1362
        int           HDrawCycles;                        // NTSC:1024        PAL:1024
        int           HBlankCycles;                       // NTSC: 340        PAL: 338
        int           ScanlineEndCycles;                  // NTSC:   4        PAL:   2
        int           FrameCycles;                        // NTSC:29829.52    PAL:35468.94
        int           FrameIrqCycles;                     // NTSC:29829.52    PAL:35468.94
        int           FrameRate;                          // NTSC:60(59.94)   PAL:50
        unsigned long FramePeriod;                        // NTSC:16.683      PAL:20.0
} NESCONFIG, *LPNESCONFIG;


typedef struct tagSPRITE
{
        BYTE Y;
        BYTE Index;
        BYTE Attribute;
        BYTE X;
} SPRITE;

extern int VROM_1K_SIZE;
extern int VROM_2K_SIZE;
extern int VROM_4K_SIZE;
extern int VROM_8K_SIZE;

extern int PROM_8K_SIZE;
extern int PROM_16K_SIZE;
extern int PROM_32K_SIZE;

extern void Mirror_type( INT type );
extern void SetVRAM_Mirror( INT type );
extern BYTE RevertByte[256];
extern BYTE RomHeader[16];
extern BYTE GRAM[0x800];
extern BYTE SRAM[0x800];
extern BYTE PRGBlock[4*PRG_BLOCK_SIZE];
extern BYTE PatternTable[4*PAT_BLOCK_SIZE];
extern BYTE NameTable[0x800];
extern BYTE BGPal[0x10];
extern BYTE SPPal[0x10];
extern BYTE SPRAM[0x100];
extern SPRITE * Sprite;
extern int PROM_8K_SIZE;
extern BYTE *CPU_MEM_BANK[8];
extern BYTE *PPU_MEM_BANK[12];
extern BYTE VRAM[4*1024];//4

void RUN_WaveNES(void);
void NES_Init(void);
void NES_Start(void);
void NES_FrameExec(void);
void LoadRomFile(char *NES_ROM);
BYTE NES_Read(WORD addr);
void NES_Write(WORD addr, BYTE val);
BYTE * NES_GetRAM(WORD addr);
BYTE NES_ReadReg(WORD addr);
void NES_WriteReg(WORD addr, BYTE val);
void SetPROM_8K_Bank(BYTE page, int bank);
void SetPROM_16K_Bank(BYTE page, int bank);
void SetPROM_32K_Bank1(int bank);
void SetPROM_32K_Bank2(int bank0, int bank1, int bank2, int bank3);
void SetVROM_1K_Bank(BYTE page, int bank);
void SetVROM_2K_Bank(BYTE page, int bank);
void SetVROM_4K_Bank(BYTE page, int bank);
void SetVROM_8K_Bank(int bank);
void SetCRAM_1K_Bank(BYTE page, int bank);
void SetCRAM_2K_Bank(BYTE page, int bank);
void SetCRAM_4K_Bank(BYTE page, int bank);
void SetCRAM_8K_Bank(int bank);
void SetVROM_8K_Bank1(int bank0, int bank1, int bank2, int bank3, int bank4, int bank5, int bank6, int bank7);
void SetNameTable_Bank(int bank0, int bank1, int bank2, int bank3);

#endif


/*
  A. iNES Format (.NES)
  ---------------------
    +--------+------+------------------------------------------+
    | Offset | Size | Content(s)                               |
    +--------+------+------------------------------------------+
    |   0    |  3   | 'NES'                                    |
    |   3    |  1   | $1A                                      |
    |   4    |  1   | 16K PRG-ROM page count                   |
    |   5    |  1   |  8K CHR-ROM page count                   |
    |   6    |  1   | ROM Control Byte #1                      |
    |        |      |   %####vTsM                              |
    |        |      |    |  ||||+- 0=Horizontal mirroring      |
    |        |      |    |  ||||   1=Vertical mirroring        |
    |        |      |    |  |||+-- 1=SRAM enabled              |
    |        |      |    |  ||+--- 1=512-byte trainer present  |
    |        |      |    |  |+---- 1=Four-screen mirroring     |
    |        |      |    |  |                                  |
    |        |      |    +--+----- Mapper # (lower 4-bits)     |
    |   7    |  1   | ROM Control Byte #2                      |
    |        |      |   %####00PV                              |
    |        |      |    |  |  |+- 1=VS-Unisystem Images       |
    |        |      |    |  |  +-- 1=PlayChoice10 Images       |
    |        |      |    |  |                                  |
    |        |      |    +--+----- Mapper # (upper 4-bits)     |
    |  8-15  |  8   | $00                                      |
    | 16-..  |      | Actual 16K PRG-ROM pages (in linear      |
    |  ...   |      | order). If a trainer exists, it precedes |
    |  ...   |      | the first PRG-ROM page.                  |
    | ..-EOF |      | CHR-ROM pages (in ascending order).      |
    +--------+------+------------------------------------------+
*/